package pp.s1184725.boppi;

import java.util.*;

import org.antlr.v4.runtime.ParserRuleContext;
import org.antlr.v4.runtime.RuleContext;
import org.antlr.v4.runtime.tree.*;

import pp.iloc.model.*;
import pp.s1184725.boppi.BasicParser.*;

public class BoppiBasicGenerator extends BasicBaseVisitor<Void> {
	private static final int ARBASESIZE = 8;
	public Program prog;
	private Annotations an;
	private int regNum;
	private final Reg tempArp = new Reg("ART"), arp = new Reg("ARP"), resultReg = new Reg("RES");

	/**
	 * Maps operator tokens to ILOC opcodes.
	 */
	private static Map<Integer, OpCode> ops = new HashMap<Integer, OpCode>() {
		private static final long serialVersionUID = 8979722313842633807L;

		{
			put(BasicLexer.AND, OpCode.and);
			put(BasicLexer.DIVIDE, OpCode.div);
			put(BasicLexer.EQ, OpCode.cmp_EQ);
			put(BasicLexer.GT, OpCode.cmp_GT);
			put(BasicLexer.GTE, OpCode.cmp_GE);
			put(BasicLexer.LEQ, OpCode.cmp_LE);
			put(BasicLexer.LT, OpCode.cmp_LT);
			put(BasicLexer.MINUS, OpCode.sub);
			put(BasicLexer.MULTIPLY, OpCode.mult);
			put(BasicLexer.NEQ, OpCode.cmp_NE);
			put(BasicLexer.PLUS, OpCode.add);
			put(BasicLexer.OR, OpCode.or);
		}
	};

	/**
	 * Creates a new generator with the given variable annotations.
	 * 
	 * @param annotations
	 *            the annotation object to use during generation
	 */
	public BoppiBasicGenerator(Annotations annotations) {
		an = annotations;
		prog = new Program();
		regNum = 0;
	}

	/**
	 * Retrieves a register for a given tree node.
	 * 
	 * @param parseTree
	 *            the tree node
	 * @return a register for the tree node
	 */
	private Reg getReg(ParseTree parseTree) {
		Reg reg = an.registers.get(parseTree);
		if (reg == null) {
			reg = new Reg("r_" + (regNum++));
			an.registers.put(parseTree, reg);
		}
		return reg;
	}

	/**
	 * Attaches the register used in the {@code source} tree node to the
	 * {@link destination} node.
	 * 
	 * @param source
	 *            the node to retrieve the register from
	 * @param destination
	 *            the node to copy the register to
	 * @return the register that was copied
	 */
	private Reg copyReg(ParseTree source, ParseTree destination) {
		Reg reg = an.registers.get(source);
		an.registers.put(destination, reg);
		return reg;
	}

	private <T> Reg getVar(ParserRuleContext node, Variable<T> var) {
		int difference = an.function.get(node).getDepth() + 1 - var.getDepth();
		if (difference > 0) {
			emit(node, OpCode.i2i, arp, tempArp);
			while (difference-- > 0)
				emit(node, OpCode.loadAI, tempArp, new Num(-4), tempArp);

			return tempArp;
		} else
			return arp;
	}

	/**
	 * Constructs an operation from the parameters and adds it to the program
	 * under construction.
	 */
	private Op emit(ParserRuleContext ctx, Label label, OpCode opCode, Operand... args) {
		Op result = new Op(label, opCode, args);
		result.setComment(ctx.getText());
		this.prog.addInstr(result);
		return result;
	}

	/**
	 * Constructs an operation from the parameters and adds it to the program
	 * under construction.
	 */
	private Op emit(ParserRuleContext ctx, OpCode opCode, Operand... args) {
		return emit(ctx, (Label) null, opCode, args);
	}

	@Override
	public Void visitChildren(RuleNode tree) {
		super.visitChildren(tree);
		if (an.registers.get(tree) == null)
			for (int i = tree.getChildCount() - 1; i >= 0; i--)
				if (tree.getChild(i) instanceof RuleContext) {
					copyReg(tree.getChild(i), tree);
					break;
				}

		return null;
	}

	@Override
	public Void visitProgram(ProgramContext ctx) {
		emit(ctx, OpCode.loadI, new Num(0), tempArp);
		emit(ctx, OpCode.loadI, new Num(0), arp);

		return visitChildren(ctx);
	}

	@Override
	public Void visitAssign(AssignContext ctx) {
		visit(ctx.singleExpr());
		copyReg(ctx.singleExpr(), ctx);

		if (SimpleType.CHAR.equals(an.types.get(ctx)))
			emit(ctx, OpCode.cstoreAI, getReg(ctx), getVar(ctx.variable(), an.variables.get(ctx.variable())),
					new Num(an.variables.get(ctx.variable()).getOffset()));
		else
			emit(ctx, OpCode.storeAI, getReg(ctx), getVar(ctx.variable(), an.variables.get(ctx.variable())),
					new Num(an.variables.get(ctx.variable()).getOffset()));

		return null;
	}

	@Override
	public Void visitCall(CallContext ctx) {
		Variable<Type> function = an.variables.get(ctx.variable());
		FunctionType parameters = ((FunctionType) function.getType());
		Reg temp = new Reg("r__");
		visit(ctx.variable());
		emit(ctx.variable(), OpCode.loadAI, getVar(ctx.variable(), an.variables.get(ctx.variable())),
				new Num(an.variables.get(ctx.variable()).getOffset()), temp);
		emit(ctx.variable(), OpCode.push, temp);

		for (int i = 0; i < ctx.expr().size(); i++) {
			visit(ctx.expr(i));

			if (SimpleType.CHAR.equals(parameters.getParameters().get(i)))
				emit(ctx.expr(i), OpCode.cpush, getReg(ctx.expr(i)));
			else
				emit(ctx.expr(i), OpCode.push, getReg(ctx.expr(i)));
		}
		emit(ctx.variable(), OpCode.addI, arp,
				new Num(((FunctionType) an.function.get(ctx).getType()).getLocalDataSize() + ARBASESIZE), tempArp);
		for (int i = ctx.expr().size() - 1; i >= 0; i--) {
			if (SimpleType.CHAR.equals(parameters.getParameters().get(i))) {
				emit(ctx.expr(i), OpCode.cpop, temp);
				emit(ctx.expr(i), OpCode.cstoreAI, temp, tempArp, new Num(parameters.getOffsets().get(i)));
			} else {
				emit(ctx.expr(i), OpCode.pop, temp);
				emit(ctx.expr(i), OpCode.storeAI, temp, tempArp, new Num(parameters.getOffsets().get(i)));
			}
		}
		emit(ctx.variable(), OpCode.storeAI, arp, tempArp, new Num(-4));
		Label label = new Label("ret" + ctx.hashCode());
		emit(ctx.variable(), OpCode.loadI, new Num(label), temp);
		emit(ctx.variable(), OpCode.storeAI, temp, tempArp, new Num(-8));
		emit(ctx, OpCode.i2i, tempArp, arp);
		emit(ctx.variable(), OpCode.pop, temp);
		emit(ctx, OpCode.jump, temp);
		emit(ctx, label, OpCode.loadAI, arp, new Num(-4), arp);

		an.registers.put(ctx, resultReg);
		return null;
	}

	@Override
	public Void visitFuncDeclare(FuncDeclareContext ctx) {
		Reg temp = new Reg("r__");
		Label skip = new Label("s" + ctx.hashCode());
		emit(ctx, OpCode.jumpI, skip);
		int targetAddress = emit(ctx, OpCode.nop).getLine();
		visit(ctx.body);
		copyReg(ctx.body, ctx);
		if (SimpleType.CHAR.equals(an.types.get(ctx.body)))
			emit(ctx, OpCode.c2c, getReg(ctx), resultReg);
		else
			emit(ctx, OpCode.i2i, getReg(ctx), resultReg);
		emit(ctx, OpCode.loadAI, arp, new Num(-8), temp);
		emit(ctx, OpCode.jump, temp);
		emit(ctx, skip, OpCode.loadI, new Num(targetAddress), temp);
		emit(ctx, OpCode.storeAI, temp, getVar(ctx, an.variables.get(ctx)), new Num(an.variables.get(ctx).getOffset()));
		return null;
	}

	@Override
	public Void visitBool(BoolContext ctx) {
		emit(ctx, OpCode.loadI, new Num(ctx.TRUE() != null ? 1 : 0), getReg(ctx));
		return null;
	}

	@Override
	public Void visitNumber(NumberContext ctx) {
		emit(ctx, OpCode.loadI, new Num(Integer.parseInt(ctx.LITERAL10().getText())), getReg(ctx));
		return null;
	}

	@Override
	public Void visitInfix1(Infix1Context ctx) {
		visitChildren(ctx);
		emit(ctx, ops.get(ctx.op.getType()), getReg(ctx.lhs), getReg(ctx.rhs), getReg(ctx));
		return null;
	}

	@Override
	public Void visitInfix2(Infix2Context ctx) {
		visitChildren(ctx);
		emit(ctx, ops.get(ctx.op.getType()), getReg(ctx.lhs), getReg(ctx.rhs), getReg(ctx));
		return null;
	}

	@Override
	public Void visitInfix3(Infix3Context ctx) {
		visitChildren(ctx);
		emit(ctx, ops.get(ctx.op.getType()), getReg(ctx.lhs), getReg(ctx.rhs), getReg(ctx));
		return null;
	}

	@Override
	public Void visitInfix4(Infix4Context ctx) {
		visitChildren(ctx);
		emit(ctx, OpCode.and, getReg(ctx.lhs), getReg(ctx.rhs), getReg(ctx));
		return null;
	}

	@Override
	public Void visitInfix5(Infix5Context ctx) {
		visitChildren(ctx);
		emit(ctx, OpCode.or, getReg(ctx.lhs), getReg(ctx.rhs), getReg(ctx));
		return null;
	}

	@Override
	public Void visitPrefix1(Prefix1Context ctx) {
		visitChildren(ctx);

		switch (ctx.op.getType()) {
		case BasicLexer.MINUS:
			emit(ctx, OpCode.rsubI, getReg(ctx.singleExpr()), new Num(0), getReg(ctx));
		}
		return null;
	}

	@Override
	public Void visitVar(VarContext ctx) {
		if (SimpleType.CHAR.equals(an.types.get(ctx)))
			emit(ctx, OpCode.cloadAI, getVar(ctx.variable(), an.variables.get(ctx.variable())),
					new Num(an.variables.get(ctx.variable()).getOffset()), getReg(ctx));
		else
			emit(ctx, OpCode.loadAI, getVar(ctx.variable(), an.variables.get(ctx.variable())),
					new Num(an.variables.get(ctx.variable()).getOffset()), getReg(ctx));

		return null;
	}

	@Override
	public Void visitChar(CharContext ctx) {
		emit(ctx, OpCode.loadI, new Num(ctx.CHAR().getText().codePointAt(1)), getReg(ctx));
		emit(ctx, OpCode.i2c, getReg(ctx), getReg(ctx));
		return null;
	}

	@Override
	public Void visitRead(ReadContext ctx) {
		for (VariableContext expr : ctx.variable()) {
			visit(expr);
			if (SimpleType.BOOL.equals(an.types.get(expr)) || SimpleType.INT.equals(an.types.get(expr))) {
				emit(ctx, OpCode.in, new Str(""), getReg(ctx));
				emit(ctx, OpCode.storeAI, getReg(ctx), getVar(expr, an.variables.get(expr)),
						new Num(an.variables.get(expr).getOffset()));
			} else if (SimpleType.CHAR.equals(an.types.get(expr))) {
				// Get input until at least 1 character
				Label getTarget = new Label("lcin_l" + ctx.hashCode());
				Label continueTarget = new Label("lcin_e" + ctx.hashCode());
				Reg countReg = new Reg("tempReg");
				emit(ctx, getTarget, OpCode.cin, new Str(""));
				emit(ctx, OpCode.pop, countReg);
				emit(ctx, OpCode.cbr, countReg, continueTarget, getTarget);

				// Get character
				emit(ctx, continueTarget, OpCode.cpop, getReg(ctx));
				emit(ctx, OpCode.cstoreAI, getReg(ctx), getVar(expr, an.variables.get(expr)),
						new Num(an.variables.get(expr).getOffset()));

				// Pop all remaining characters
				Label loopTarget = new Label("lcpop_l" + ctx.hashCode());
				Label iterTarget = new Label("lcpop_c" + ctx.hashCode());
				Label stopTarget = new Label("lcpop_e" + ctx.hashCode());
				Reg tempReg2 = new Reg("tempReg2");
				emit(ctx, loopTarget, OpCode.subI, countReg, new Num(1), countReg);
				emit(ctx, OpCode.cbr, countReg, iterTarget, stopTarget);
				emit(ctx, iterTarget, OpCode.cpop, tempReg2);
				emit(ctx, OpCode.jumpI, loopTarget);
				emit(ctx, stopTarget, OpCode.nop);
			} else {
				emit(ctx, OpCode.out, new Str("Unknown type: "), an.registers.get(expr));
			}
		}
		return null;
	}

	@Override
	public Void visitIf(IfContext ctx) {
		Label toTrue = new Label("if_t" + ctx.hashCode());
		Label toFalse = new Label("if_f" + ctx.hashCode());
		Label toEnd = new Label("if_e" + ctx.hashCode());

		visit(ctx.cond);
		if (ctx.onFalse == null) {
			emit(ctx, OpCode.cbr, getReg(ctx.cond), toTrue, toEnd);

			emit(ctx, toTrue, OpCode.nop);
			visit(ctx.onTrue);
		} else {
			emit(ctx, OpCode.cbr, getReg(ctx.cond), toTrue, toFalse);

			emit(ctx, toTrue, OpCode.nop);
			visit(ctx.onTrue);
			if (an.types.get(ctx) != SimpleType.VOID)
				emit(ctx, OpCode.i2i, getReg(ctx.onTrue), getReg(ctx));
			emit(ctx, OpCode.jumpI, toEnd);

			emit(ctx, toFalse, OpCode.nop);
			visit(ctx.onFalse);
			if (an.types.get(ctx) != SimpleType.VOID)
				emit(ctx, OpCode.i2i, getReg(ctx.onFalse), getReg(ctx));

		}
		emit(ctx, toEnd, OpCode.nop);

		return null;
	}

	@Override
	public Void visitWhile(WhileContext ctx) {
		Label toLoop = new Label("while_t" + ctx.hashCode());
		Label toCond = new Label("while_f" + ctx.hashCode());
		Label toEnd = new Label("while_e" + ctx.hashCode());

		emit(ctx, OpCode.jumpI, toCond);
		emit(ctx, toLoop, OpCode.nop);
		visit(ctx.onTrue);
		emit(ctx, toCond, OpCode.nop);
		visit(ctx.cond);
		emit(ctx, OpCode.cbr, getReg(ctx.cond), toLoop, toEnd);
		emit(ctx, toEnd, OpCode.nop);

		return null;
	}

	@Override
	public Void visitWrite(WriteContext ctx) {
		for (ExprContext expr : ctx.expr()) {
			visit(expr);
			if (SimpleType.BOOL.equals(an.types.get(expr)) || SimpleType.INT.equals(an.types.get(expr))) {
				emit(ctx, OpCode.out, new Str(""), getReg(expr));
			} else if (SimpleType.CHAR.equals(an.types.get(expr))) {
				Reg tempReg = new Reg("tempReg");
				emit(ctx, OpCode.cpush, getReg(expr));
				emit(ctx, OpCode.loadI, new Num(1), tempReg);
				emit(ctx, OpCode.push, tempReg);
				emit(ctx, OpCode.cout, new Str(""));
			} else {
				emit(ctx, OpCode.out, new Str("Unknown type: "), getReg(expr));
			}
		}
		copyReg(ctx.expr(0), ctx);
		return null;
	}
}
